# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2021 by the Free Software Foundation, Inc.
#
# This file is part of Django-Mailman.
#
# Django-Mailman is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Django-Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Django-Mailman.  If not, see <http://www.gnu.org/licenses/>.


from django.test import RequestFactory, TestCase, override_settings

from django_mailman3.context_processors import common


class TestContextProcessors(TestCase):

    def setUp(self):
        self.factory = RequestFactory()

    @override_settings(LOGIN_URL='/login', LOGOUT_URL='/logout',
                       ALLOWED_HOSTS=['example.com'])
    def test_common_with_all_vars(self, *args, **kwargs):
        request = self.factory.get('/', HTTP_HOST='example.com')
        context = common(request)
        self.assertEqual(context['site_name'], 'example.com')
        self.assertEqual(context['LOGIN_URL'], '/login')
        self.assertEqual(context['LOGOUT_URL'], '/logout')



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2021 by the Free Software Foundation, Inc.
#
# This file is part of Django-Mailman.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Django-Mailman.  If not, see <http://www.gnu.org/licenses/>.
#

from django.forms import ChoiceField, DateField, EmailField, Form
from django.forms.widgets import CheckboxInput
from django.template import Context, Template
from django.test import SimpleTestCase

from django_mailman3.templatetags.bootstrap_tags import (
    fieldtype, fieldtype_is, is_checkbox)


class TestForm(Form):
    datetime = DateField()
    email = EmailField()
    choices = ChoiceField(choices=('yes', 'no'), widget=CheckboxInput)


class TestBootstrapTags(SimpleTestCase):

    def setUp(self):
        self.form = TestForm()

    def test_add_form_control(self):
        # Given a field the field class has 'form-control' in it.
        context = Context({'form': self.form})
        template = Template(
            '{% load bootstrap_tags %}'
            '{% for field in form.visible_fields %}'
            '{{ field | add_form_control }} <br />'
            '{% endfor %}')
        rendered_template = template.render(context=context)
        datetime_field = 'class="form-control"'
        self.assertTrue(datetime_field in rendered_template)

    def test_fieldtype(self):
        # Fieldtype takes a bounded formfield and returns it's type.
        fields = [fieldtype(field) for field in self.form.visible_fields()]
        self.assertEqual(sorted(fields),
                         ['CheckboxInput', 'DateInput', 'EmailInput'])
        # If the wrong instance of the field is passed.
        with self.assertRaises(AssertionError):
            fieldtype(self.form.fields['email'])

    def test_fieldtype_is(self):
        # Fields should return true for all of it's parent classes in
        # the hierarchy.
        for field in self.form.visible_fields():
            if field.name == 'datetime':
                self.assertTrue(fieldtype_is(field, 'DateInput'))
                self.assertTrue(fieldtype_is(field, 'DateTimeBaseInput'))
                self.assertTrue(fieldtype_is(field, 'TextInput'))
                self.assertFalse(fieldtype_is(field, 'RadioSelect'))
            if field.name == 'email':
                self.assertTrue(fieldtype_is(field, 'EmailInput'))
                self.assertTrue(fieldtype_is(field, 'Input'))
            if field.name == 'choices':
                self.assertTrue(fieldtype_is(field, 'CheckboxInput'))

    def test_is_checkbox(self):
        # There should a checkbox in our current field.
        fields = list(self.form.visible_fields())
        self.assertEqual(sorted(map(is_checkbox, fields)),
                         [False, False, True])
        for field in self.form.visible_fields():
            if field.name == 'choices':
                self.assertTrue(is_checkbox(field))

    def test_boostrap_form(self):
        # Test that bootstrap_form renders the form without errors.
        context = Context({'title': 'My Title',
                           'form': self.form,
                           'button': "Yes"})
        # Test the template is rendered.
        template = Template(
            '{% load bootstrap_tags %}'
            '<form> {% bootstrap_form form %}')
        rendered_template = template.render(context=context)
        label_field = '''
<label class="control-label font-weight-bold" for="id_datetime">
    Datetime
</label>'''
        self.assertInHTML(label_field, rendered_template)
        # Now render the form with a submit buttom
        template = Template(
            '{% load bootstrap_tags %}'
            '<form> {% bootstrap_form form=form button=button %}')
        rendered_template = template.render(context=context)
        button_html = '''
<div class="form-group">
    <button class="btn btn-primary" type="submit">Yes</button>
</div>
'''
        self.assertInHTML(button_html, rendered_template)

    def test_boostrap_form_horizontal(self):
        # Test that bootstrap_form_horizontal renders the form and includes the
        # submit button.
        context = Context({'title': 'My Title',
                           'form': self.form,
                           'button': "Yes"})
        template = Template(
            '{% load bootstrap_tags %}'
            '<form> {% bootstrap_form_horizontal form=form button=button%}')
        rendered_template = template.render(context=context)
        self.assertInHTML(
            '<button class="btn btn-primary" type="submit">Yes</button>',
            rendered_template)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2021 by the Free Software Foundation, Inc.
#
# This file is part of Django-Mailman.
#
# Django-Mailman is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Django-Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Django-Mailman.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#


from django.core.paginator import EmptyPage, InvalidPage, Paginator
from django.http import Http404
from django.utils.functional import cached_property


class MailmanPaginator(Paginator):
    """
    Subclass of Django's Paginator that works with MailmanClient's 'get_*_page'
    functions. Use the function reference as the first argument::

        MailmanPaginator(get_member_page, 25)

    """

    def __init__(self, function, per_page, **kwargs):
        self.function = function
        super(MailmanPaginator, self).__init__(None, per_page, **kwargs)

    def page(self, number):
        """
        Returns a Page object for the given 1-based page number.
        """
        number = self.validate_number(number)
        result = self.function(count=self.per_page, page=number)
        return self._get_page(result, number, self)

    @cached_property
    def count(self):
        """
        Returns the total number of objects, across all pages.
        """
        # See the rest/docs/collections.rst in the Mailman source tree.
        return self.function(count=0, page=1).total_size


def paginate(objects=None, page_num=None, results_per_page=None,
             max_page_range=10, paginator_class=Paginator):
    try:
        page_num = int(page_num)
    except (ValueError, TypeError):
        page_num = 1
    try:
        results_per_page = int(results_per_page)
    except (ValueError, TypeError):
        results_per_page = 10
    paginator = paginator_class(objects, results_per_page)
    try:
        objects = paginator.page(page_num)
    except EmptyPage:
        # If page is out of range (e.g. 9999), deliver last page of results.
        objects = paginator.page(paginator.num_pages)
    except InvalidPage:
        # This never happens with Django's Paginator, but just in case.
        raise Http404("No such page of results!")
    # Calculate the displayed page range
    if paginator.num_pages > max_page_range:
        paginator.page_range_ellipsis = [1]
        subrange_lower = page_num - int(max_page_range / 2 - 2)
        if subrange_lower > 3:
            paginator.page_range_ellipsis.append("...")
        else:
            subrange_lower = 2
        paginator.page_range_ellipsis.extend(
            list(range(subrange_lower, page_num)))
        if page_num != 1 and page_num != paginator.num_pages:
            paginator.page_range_ellipsis.append(page_num)
        subrange_upper = page_num + int(max_page_range / 2 - 2)
        if subrange_upper >= paginator.num_pages - 2:
            subrange_upper = paginator.num_pages - 1
        paginator.page_range_ellipsis.extend(
            list(range(page_num+1, subrange_upper+1)))
        if subrange_upper < paginator.num_pages - 2:
            paginator.page_range_ellipsis.append("...")
        paginator.page_range_ellipsis.append(paginator.num_pages)
    else:
        paginator.page_range_ellipsis = paginator.page_range
    return objects



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2021 by the Free Software Foundation, Inc.
#
# This file is part of Django-Mailman.
#
# Django-Mailman is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Django-Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Django-Mailman.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#


from unittest.mock import Mock

from django.contrib.auth.models import User
from django.urls import reverse

from django_mailman3.models import Profile
from django_mailman3.tests.utils import TestCase


class ProfileViewTestCase(TestCase):

    def setUp(self):
        self.mm_user = Mock()
        self.mm_user.user_id = "dummy"
        self.mm_user.addresses = [Mock()]
        self.mailman_client.get_user.side_effect = lambda e: self.mm_user
        self.user = User.objects.create_user(
            'testuser', 'test@example.com', 'testPass',
            first_name="firstname", last_name="lastname",
            )
        self.client.login(username='testuser', password='testPass')

    def test_get_page(self):
        response = self.client.get(reverse('mm_user_profile'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(list(response.context['other_addresses']), [])

    def test_change_display_name(self):
        # We create a Mailman user, from the django user object.
        response = self.client.post(
            reverse('mm_user_profile'), {
                'username': self.user.username,
                'first_name': 'test-first-name',
                'last_name': 'test-last-name',
                'timezone': 'Europe/Paris'})
        self.assertRedirects(response, reverse('mm_user_profile'))
        # The user's first and last name must be updated.
        user = User.objects.get(username='testuser')
        self.assertEqual(user.first_name, 'test-first-name')
        self.assertEqual(user.last_name, 'test-last-name')
        # The profile must have been created and populated.
        profile = Profile.objects.get(user=user)
        self.assertIsNotNone(profile)
        self.assertEqual(profile.timezone, 'Europe/Paris')
        # The Mailman user's display name must have been updated.
        self.assertEqual(
            self.mm_user.display_name, 'test-first-name test-last-name')

    def test_username_unique(self):
        # Usernames should be unique.
        User.objects.create_user(
            'otheruser', 'other@example.com', 'testPass',
            )
        response = self.client.post(
            reverse('mm_user_profile'), {
                'username': 'otheruser',
                'first_name': 'first-name',
                'last_name': 'last-name',
                'timezone': 'Europe/Paris'})
        self.assertEqual(response.status_code, 200)
        self.assertIn('username', response.context["form"].errors)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2021 by the Free Software Foundation, Inc.
#
# This file is part of Django-Mailman.
#
# Django-Mailman is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Django-Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Django-Mailman.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#


from django.conf import settings
from django.contrib.sites.models import Site
from django.db import models

import pytz


#: A list of common timezones as options for a user to choose their own.
TIMEZONES = sorted([(tz, tz) for tz in pytz.common_timezones])


class Profile(models.Model):
    user = models.OneToOneField(settings.AUTH_USER_MODEL,
                                related_name="mailman_profile",
                                on_delete=models.CASCADE)
    timezone = models.CharField(max_length=100, choices=TIMEZONES, default="")

    def __str__(self):
        return '<Mailman profile for %s>' % self.user.username


class MailDomain(models.Model):
    site = models.ForeignKey(Site, related_name="mailman_domains",
                             on_delete=models.CASCADE)
    mail_domain = models.CharField(max_length=255, db_index=True, unique=True)

    def __str__(self):
        return '<Mailman domain %s>' % self.mail_domain



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2021 by the Free Software Foundation, Inc.
#
# This file is part of Django-Mailman3.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Django-Mailman3.  If not, see <http://www.gnu.org/licenses/>.


from unittest.mock import Mock, patch

from django.test import RequestFactory, TestCase
from django.urls import reverse

from openid.consumer import consumer

from django_mailman3.lib.auth.fedora.provider import (
    FedoraAccount, FedoraProvider)


class TestFedoraAccount(TestCase):
    """
    Test FedoraAccount social account.
    """

    def setUp(self):
        self.account = FedoraAccount('')

    def test_self_brand(self):
        self.assertEqual(self.account.get_brand(),
                         {'id': 'fedora', 'name': 'Fedora'})


class TestFedoraProvider(TestCase):
    """
    Test FedoraProvider openid authentication.
    """

    @patch('allauth.socialaccount.providers.openid.views._openid_consumer')
    def setUp(self, consumer_mock):
        self.factory = RequestFactory()
        client = Mock()
        complete = Mock()
        consumer_mock.return_value = client
        client.complete = complete
        self.complete_response = Mock()
        complete.return_value = self.complete_response
        self.complete_response.status = consumer.SUCCESS
        self.complete_response.identity_url = 'http://bob.id.fedoraproject.org'

    def test_get_login_url(self):
        req = self.factory.get('/')
        login_url = FedoraProvider(req).get_login_url(req)
        self.assertEqual(login_url, reverse('fedora_login'))
        login_url = FedoraProvider(req).get_login_url(req, query1='value1')
        new_url = reverse('fedora_login') + '?query1=value1'
        self.assertEqual(login_url, new_url)

    def test_extract_username(self):
        req = self.factory.get('/')
        username = FedoraProvider(req).extract_username(self.complete_response)
        self.assertEqual(username, 'bob')

    def test_extract_commmon_fields(self):
        # This this test we patch super(FedoraProvider, provider) so that it's
        # complicated extract_common_fields method is not called. It is too
        # complicated to try to find the return value and built a mock that
        # satisfies the request. The only relevant output from this we need is
        # a dictionary which is then extended by our FedoraProvider class.
        mock_parent = Mock()
        mock_parent.extract_common_fields.return_value = {}
        # To patch the superclass, we patch the super() builtin.
        with patch('builtins.super') as super_mock:
            super_mock.return_value = mock_parent
            req = self.factory.get('/')
            provider = FedoraProvider(req)
            res = provider.extract_common_fields(self.complete_response)
        self.assertEqual(res, {'username': 'bob'})

    def test_extract_email_addresses(self):
        with patch('django_mailman3.lib.auth.fedora.provider'
                   '.get_email_from_response') as email_mock:
            email_mock.return_value = 'testuser@example.com'
            req = self.factory.get('/')
            emails = FedoraProvider(req).extract_email_addresses(
                self.complete_response)
        self.assertEqual(len(emails), 2)
        self.assertEqual(sorted([x.email for x in emails]),
                         ['bob@fedoraproject.org', 'testuser@example.com'])



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2021 by the Free Software Foundation, Inc.
#
# This file is part of Postorius.
#
# Postorius is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Postorius is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Postorius.  If not, see <http://www.gnu.org/licenses/>.

from django import template
from django.forms.boundfield import BoundField


register = template.Library()


@register.filter(name='add_form_control')
def add_form_control(field):
    return field.as_widget(attrs={'class': 'form-control'})


@register.filter('fieldtype')
def fieldtype(field):
    """
    Return the name of the field's class.

    :name field: Field should a given field.
    :type field: django.forms.boundfield.BoundField
    """
    assert isinstance(field, BoundField)
    return field.field.widget.__class__.__name__


@register.filter('fieldtype_is')
def fieldtype_is(field, widget_class):
    """
    Given a field and a widget class, check if the field is a subclass of the
    widget class.

    :name field: Field should a given field.
    :type field: django.forms.boundfield.BoundField
    :name widget_class: Django form widget class.
    :type widget_class: django.forms.widgets.Widget
    """
    assert isinstance(field, BoundField)
    return widget_class in [
        parent.__name__ for parent in
        field.field.widget.__class__.__mro__]


@register.filter('is_checkbox')
def is_checkbox(field):
    """
    Given the field, check if it is a checkbox field.

    :name field: Field should a given field.
    :type field: django.forms.boundfield.BoundField
    """
    return field.field.widget.__class__.__name__ in (
        'CheckboxInput', 'CheckboxSelectMultiple')


@register.inclusion_tag('django_mailman3/bootstrap/form.html',
                        takes_context=True)
def bootstrap_form(context, form, button=None):
    """
    Given a form, renders it using the bootstrapped template.
    """
    return dict(
        form=form,
        button=button,
        )


@register.inclusion_tag('django_mailman3/bootstrap/form-horizontal.html',
                        takes_context=True)
def bootstrap_form_horizontal(
        context, form, size_left=2, size_right=8, button=None,
        fold_class='md'):
    """
    Given a form object, renders the form horizontally using the bootstrap
    template.
    """
    return dict(
        form=form,
        size_left=size_left,
        size_right=size_right,
        button=button,
        fold_class=fold_class,
        )



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2021 by the Free Software Foundation, Inc.
#
# This file is part of Django-Mailman.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Django-Mailman.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from allauth.socialaccount import providers
from allauth.socialaccount.app_settings import QUERY_EMAIL
from allauth.socialaccount.helpers import (
    complete_social_login, render_authentication_error)
from allauth.socialaccount.models import SocialLogin
from allauth.socialaccount.providers.base import AuthError
from allauth.socialaccount.providers.openid.forms import LoginForm
from allauth.socialaccount.providers.openid.utils import (
    AXAttributes, SRegFields)
from allauth.socialaccount.providers.openid.views import _openid_consumer
from openid.consumer import consumer
from openid.consumer.discover import DiscoveryFailure
from openid.extensions.ax import AttrInfo, FetchRequest
from openid.extensions.sreg import SRegRequest

from .provider import FedoraProvider


class LoginView(View):

    form_class = LoginForm
    template_name = 'openid/login.html'
    provider = FedoraProvider
    callback_view = 'fedora_callback'

    def get(self, request, *args, **kwargs):
        if 'openid' in request.GET or self.provider.endpoint:
            return self.post(request, *args, **kwargs)
        form = LoginForm(initial={'next': request.GET.get('next'),
                                  'process': request.GET.get('process')})
        return render(request, self.template_name, {'form': form})

    def post(self, request, *args, **kwargs):
        data = dict(list(request.GET.items()) + list(request.POST.items()))
        if self.provider.endpoint:
            data['openid'] = self.provider.endpoint
        form = LoginForm(data)
        if form.is_valid():
            client = _openid_consumer(request)
            try:
                auth_request = client.begin(form.cleaned_data['openid'])
                if QUERY_EMAIL:
                    sreg = SRegRequest()
                    for name in SRegFields:
                        sreg.requestField(field_name=name,
                                          required=True)
                    auth_request.addExtension(sreg)
                    ax = FetchRequest()
                    for name in AXAttributes:
                        ax.add(AttrInfo(name,
                                        required=True))
                    auth_request.addExtension(ax)
                callback_url = reverse(self.callback_view)
                SocialLogin.stash_state(request)
                redirect_url = auth_request.redirectURL(
                    request.build_absolute_uri('/'),
                    request.build_absolute_uri(callback_url))
                return HttpResponseRedirect(redirect_url)
            # UnicodeDecodeError:
            # see https://github.com/necaris/python3-openid/issues/1
            except (UnicodeDecodeError, DiscoveryFailure) as e:
                if request.method == 'POST':
                    form._errors["openid"] = form.error_class([e])
                else:
                    return render_authentication_error(
                        request, self.provider.id, exception=e)
        return render(request, self.template_name, {'form': form})


class CallbackView(View):

    provider = FedoraProvider

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        client = _openid_consumer(request)
        response = client.complete(
            dict(list(request.GET.items()) + list(request.POST.items())),
            request.build_absolute_uri(request.path))
        if response.status == consumer.SUCCESS:
            login = providers.registry \
                .by_id(self.provider.id) \
                .sociallogin_from_response(request, response)
            login.state = SocialLogin.unstash_state(request)
            ret = complete_social_login(request, login)
        else:
            if response.status == consumer.CANCEL:
                error = AuthError.CANCELLED
            else:
                error = AuthError.UNKNOWN
            ret = render_authentication_error(
                request,
                self.provider.id,
                error=error)
        return ret



# -*- coding: utf-8 -*-
# Copyright (C) 2020-2021 by the Free Software Foundation, Inc.
#
# This file is part of Django-Mailman.
#
# Django-Mailman3 is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Django-Mailman3 is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Django-Mailman.  If not, see <http://www.gnu.org/licenses/>.

from unittest.mock import Mock

from django.contrib.auth.models import User

from allauth.account.models import EmailAddress

from django_mailman3.lib.user import get_django_user
from django_mailman3.tests.utils import TestCase


class TestGetDjangUser(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            'testuser', 'test@example.com', 'testPass')
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True)
        EmailAddress.objects.create(
            user=self.user, email='secondemail@example.com', verified=True)
        EmailAddress.objects.create(
            user=self.user, email='thirdemail@example.com', verified=True)

    def test_get_user_with_email(self):
        # Test we can get the user from Member objects.
        for email in (
                'test@example.com', 'secondemail@example.com',
                'thirdemail@example.com'):
            member = Mock()
            member.address = Mock(email=email)
            user = get_django_user(member)
            self.assertEqual(user, self.user)

    def test_missing_user(self):
        # test that we don't raise exception when the user doesn't exist.
        member = Mock()
        member.address = Mock(email='example@example.com')
        self.assertIsNone(get_django_user(member))



"""
Django settings for testing Django-Mailman.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.8/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os


BASE_DIR = os.path.dirname(os.path.abspath(__file__))

TESTING = True

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.8/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'change-this-on-your-production-server'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False

ADMINS = ()

# Hosts/domain names that are valid for this site; required if DEBUG is False
# See https://docs.djangoproject.com/en/1.8/ref/settings/#allowed-hosts
ALLOWED_HOSTS = []

# Mailman API credentials
MAILMAN_REST_API_URL = 'http://localhost:8001'
MAILMAN_REST_API_USER = 'restadmin'
MAILMAN_REST_API_PASS = 'restpass'

# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django_mailman3',
    'allauth',
    'allauth.account',
    'allauth.socialaccount',
    'allauth.socialaccount.providers.openid',
    'django_gravatar'
)


MIDDLEWARE = (
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)


ROOT_URLCONF = 'django_mailman3.tests.urls_test'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.template.context_processors.csrf',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# WSGI_APPLICATION = 'django_mailman3.example_project.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.8/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': '/dummy/django_mailman3.db',
    }
}


# Internationalization
# https://docs.djangoproject.com/en/1.8/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.8/howto/static-files/


# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/var/www/example.com/static/"
# STATIC_ROOT = ''
STATIC_ROOT = BASE_DIR + '/static/'

# URL prefix for static files.
# Example: "http://example.com/static/", "http://static.example.com/"
STATIC_URL = '/static/'

# Additional locations of static files
STATICFILES_DIRS = (
    # Put strings here, like "/home/html/static" or "C:/www/django/static".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
    # BASE_DIR + '/static/',
)


# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error when DEBUG=False.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'verbose',
        },
    },
    'loggers': {
        'django': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True,
        },
        'django_mailman3': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': True,
        },
    },
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(module)s: %(message)s'
        },
    },
    'root': {
        'handlers': ['console'],
        'level': 'INFO',
    },
}


# Use SSL when logged in. You need to enable the SSLRedirect middleware for
# this feature to work.
# USE_SSL = True

# Enable caching to make sure the unit tests run when it's on.
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
    }
}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2021 by the Free Software Foundation, Inc.
#
# This file is part of Django-Mailman.
#
# Django-Mailman is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Django-Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Django-Mailman.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#


# https://docs.djangoproject.com/en/dev/topics/i18n/timezones/

from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone


class TimezoneMiddleware(object):

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        if not request.user.is_authenticated:
            return response
        try:
            profile = request.user.mailman_profile
        except ObjectDoesNotExist:
            return response
        if profile.timezone:
            timezone.activate(profile.timezone)
        return response


class PaginationMiddleware(object):
    """
    Inserts a variable representing the current page onto the request object if
    it exists in either **GET** or **POST** portions of the request.
    """
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        try:
            request.page = int(request.REQUEST['page'])
        except (KeyError, ValueError, TypeError):
            request.page = 1
        return self.get_response(request)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2021 by the Free Software Foundation, Inc.
#
# This file is part of Django-Mailman.
#
# Django-Mailman is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Django-Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Django-Mailman.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from unittest.mock import Mock, call, patch
from urllib.error import HTTPError

from django.contrib.auth.models import User
from django.core.cache import cache
from django.db import IntegrityError
from django.test import override_settings

from allauth.account.models import EmailAddress

from django_mailman3.lib import mailman
from django_mailman3.tests.utils import (
    FakeMMAddress, FakeMMAddressList, FakeMMList, FakeMMMember, TestCase)


class GetMailmanUserTestCase(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            'testuser', 'test@example.com', 'testPass')
        self.mm_user = Mock()
        self.mm_user.user_id = "dummy"
        self.mm_user.addresses = [Mock()]
        self.mailman_client.get_user.side_effect = lambda e: self.mm_user

    def test_get_user(self):
        mm_user = mailman.get_mailman_user(self.user)
        self.assertIs(mm_user, self.mm_user)

    @patch('django_mailman3.lib.mailman.get_mailman_user',
           return_value=None)
    def test_get_nonexistent_user(self, mock_method):
        # Test create user fails and get_mailman_user returns None.
        mm_user = mailman.get_mailman_user(self.mm_user)
        self.assertIsNone(mm_user)
        mock_method.assert_called_with(self.mm_user)

    def test_create_user(self):
        self.mailman_client.get_user.side_effect = \
            HTTPError(None, 404, None, None, None)
        new_mm_user = Mock()
        new_mm_user.user_id = "dummy2"
        self.mailman_client.create_user.side_effect = lambda e, n: new_mm_user
        mm_user = mailman.get_mailman_user(self.user)
        self.assertEqual(
            self.mailman_client.create_user.call_args_list,
            [call(self.user.email, self.user.get_full_name())])
        self.assertIs(mm_user, new_mm_user)

    def test_connection_failed(self):
        self.mailman_client.get_user.side_effect = \
            HTTPError(None, 500, None, None, None)
        with patch('django_mailman3.lib.mailman.logger') as mock_logging:
            mm_user = mailman.get_mailman_user(self.user)
        # Make sure that the returned user is None.
        self.assertIsNone(mm_user)
        # Make sure the failure is logged.
        mock_logging.warning.assert_called()

    def test_get_user_id(self):
        mm_user_id = mailman.get_mailman_user_id(self.user)
        self.assertEqual(mm_user_id, "dummy")

    def test_bad_cached_user_id(self):
        # Test that we lookup a user via email if the cached user_id is invalid
        # and doesn't return a valid user.
        new_mm_user = Mock()
        new_mm_user.user_id = "dummy3"

        def side_effect(user_id):
            if user_id == '4':
                raise HTTPError(
                    url='', code=404, msg='No user with id 4',
                    hdrs=None, fp=None)
            elif user_id == 'test@example.com':
                return new_mm_user
            return None

        self.mailman_client.get_user.side_effect = side_effect
        # Now, also set integer user_id 4 in the cache.
        cache.set("User:{0}:mailman_user_id".format(self.user.id), '4', None)

        mm_user = mailman.get_mailman_user(self.user)
        self.assertIsNotNone(mm_user)
        self.assertEqual(mm_user, new_mm_user)

        expected = [call('4'), call('test@example.com')]
        self.mailman_client.get_user.call_args_list == expected


class AddUserToMailmanTestCase(TestCase):

    def setUp(self):
        self.ml = FakeMMList("list@example.com")
        self.mailman_client.get_list.side_effect = lambda n: self.ml
        self.ml.get_member = Mock()
        self.ml.subscribe = Mock()
        self.user = User.objects.create_user(
            'testuser', 'test@example.com', 'testPass')
        self.mm_user = Mock()
        self.mm_user.user_id = "dummy"
        self.mm_user.addresses = []
        self.mailman_client.get_user.side_effect = lambda e: self.mm_user
        self.mm_addresses = {}

    def _get_or_add_address(self, email, **kw):
        try:
            mm_addr = self.mm_addresses[email]
        except KeyError:
            mm_addr = Mock()
            mm_addr.email = email
            mm_addr.verified_on = None
            self.mm_addresses[email] = mm_addr
        return mm_addr

    def test_primary_address_unverified(self):
        self.mailman_client.get_address.side_effect = self._get_or_add_address
        self.mm_user.addresses = ["test@example.com"]
        mailman.add_address_to_mailman_user(self.user, "test@example.com")
        self.mm_addresses['test@example.com'].verify.assert_called_with()
        self.assertFalse(self.mm_user.add_address.called)

    def test_addresses_verified(self):
        self.mm_user.add_address.side_effect = self._get_or_add_address
        mailman.add_address_to_mailman_user(self.user, "secondary@example.com")
        self.mm_user.add_address.assert_called_with(
            "secondary@example.com", absorb_existing=True)
        self.mm_addresses['secondary@example.com'].verify.assert_called_with()

    def test_existing_address_but_not_verified(self):
        # The secondary address exists but is not verified
        self.mailman_client.get_address.side_effect = self._get_or_add_address
        secondary_address = Mock()
        secondary_address.email = "secondary@example.com"
        secondary_address.verified_on = None
        secondary_address.__str__ = lambda self: self.email
        self.mm_user.addresses.append(secondary_address)
        self.mm_addresses["secondary@example.com"] = secondary_address
        mailman.add_address_to_mailman_user(self.user, "secondary@example.com")
        # The secondary address must only have been verified.
        self.assertFalse(self.mm_user.add_address.called)
        secondary_address.verify.assert_called_with()

    @patch('django_mailman3.lib.mailman.logger')
    def test_add_new_address(self, mock_log):
        secondary_address = Mock()
        secondary_address.email = "secondary@example.com"
        with patch('django_mailman3.lib.mailman.get_mailman_user',
                   return_value=None):
            # If the user does not exist, function return and logs the result.
            mailman.add_address_to_mailman_user(self.user, secondary_address)
            mock_log.info.assert_called()


class SyncEmailAddressesTestCase(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            'testuser', 'test@example.com', 'testPass')
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True)
        self.mm_user = Mock()
        self.mm_user.user_id = "dummy"
        self.mm_user.addresses = FakeMMAddressList()
        self.mailman_client.get_user.side_effect = lambda e: self.mm_user

    def test_sync_django(self):
        # Addresses in Django are pushed to Mailman, verified bits are synced.
        EmailAddress.objects.create(
            user=self.user, email='another@example.com', verified=True)
        not_verified = FakeMMAddress('another@example.com', verified=False)
        self.mm_user.addresses.append(not_verified)
        mailman.sync_email_addresses(self.user)
        # The missing address must have been added
        self.assertEqual(
            self.mm_user.add_address.call_args_list,
            [call('test@example.com', absorb_existing=True)])
        # The unverified address must have been set verified
        self.assertTrue(not_verified.verified)

    def test_sync_mailman(self):
        # Addresses in Mailman are pushed to Django, verified bits are synced.
        self.mm_user.addresses.extend([
            FakeMMAddress('another@example.com', verified=True),
            FakeMMAddress('not-verified@example.com', verified=False),
            FakeMMAddress('yet-another@example.com', verified=True),
            ])
        EmailAddress.objects.create(
            user=self.user, email='yet-another@example.com', verified=False)
        mailman.sync_email_addresses(self.user)
        self.assertEqual(
            list(EmailAddress.objects.filter(
                 user=self.user, verified=True).order_by("email").values_list(
                 "email", flat=True)),
            ['another@example.com', self.user.email,
             'yet-another@example.com'])

    def test_user_conflict(self):
        # A user with two email addresses in Mailman is split in two Django
        # users.
        user2 = User.objects.create_user(
            'testuser2', 'another@example.com', 'testPass')
        EmailAddress.objects.create(
            user=user2, email=user2.email, verified=True)
        self.mm_user.addresses.extend([
            FakeMMAddress('test@example.com', verified=True),
            FakeMMAddress('another@example.com', verified=True),
            ])
        try:
            mailman.sync_email_addresses(self.user)
        except IntegrityError as e:
            self.fail(e)
        self.assertEqual(
            list(EmailAddress.objects.filter(
                user=self.user).values_list("email", flat=True)),
            [self.user.email])
        self.assertEqual(
            list(EmailAddress.objects.filter(
                user=user2).values_list("email", flat=True)),
            [user2.email])


class GetSubscriptionsTestCase(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            'testuser', 'test@example.com', 'testPass')
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True)
        self.mm_user = Mock()
        self.mm_user.user_id = "dummy"
        self.mm_user.addresses = FakeMMAddressList()
        self.mailman_client.get_user.side_effect = lambda e: self.mm_user

    def test_get_subscriptions(self):
        fake_member = FakeMMMember("list.example.com", "test@example.com")
        self.mm_user.subscriptions = [fake_member]
        self.assertEqual(
            mailman.get_subscriptions(self.user),
            {"list.example.com": "test@example.com"})

    def test_get_subscriptions_no_user(self):
        with patch('django_mailman3.lib.mailman.get_mailman_user') as gmu:
            gmu.return_value = None
            self.assertEqual(mailman.get_subscriptions(self.user), {})

    def test_get_subscriptions_nonmember(self):
        fake_member = FakeMMMember(
            "list.example.com", "test@example.com", role="nonmember")
        self.mm_user.subscriptions = [fake_member]
        self.assertEqual(mailman.get_subscriptions(self.user), {})


class UpdatePreferredAddressTestCase(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            'testuser', 'test@example.com', 'testPass')
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True)
        self.mm_user = Mock()
        self.mm_user.user_id = "dummy"
        self.mm_user.addresses = FakeMMAddressList()
        self.mailman_client.get_user.side_effect = lambda e: self.mm_user

    def test_update_preferred_address_updates_mailman(self):
        """Updating preferred address in Django updates Core."""
        self.mm_user.addresses.extend([
            FakeMMAddress('another@example.com', verified=False),
            FakeMMAddress('verified@example.com', verified=True),
            ])
        email2 = EmailAddress.objects.create(
            user=self.user, email='another@example.com', verified=False)
        email2.set_as_primary()
        mailman.update_preferred_address(self.user, email2)
        # A non-verified address should never set the preferred address.
        self.assertNotEqual(
            self.mm_user.preferred_address, 'another@example.com')
        email3 = EmailAddress.objects.create(
            user=self.user, email='verified@example.com', verified=True)
        email3.set_as_primary()
        mailman.update_preferred_address(self.user, email3)
        self.assertEqual(
            self.mm_user.preferred_address, 'verified@example.com')

    def test_update_preferred_address_adds_address(self):
        """Setting a preferred address which isn't already created in Core."""
        # Test that we can set an address which isn't already associated with
        # the user in Mailman Core.
        email4 = EmailAddress.objects.create(
            user=self.user, email='not-added@example.com', verified=True)
        email4.set_as_primary()
        mailman.update_preferred_address(self.user, email4)
        self.assertEqual(
            self.mm_user.preferred_address, 'not-added@example.com')
        self.mm_user.add_address.assert_called_once_with(
            'not-added@example.com', absorb_existing=True)


class TestGetMailmanclient(TestCase):

    @override_settings(DEBUG=True)
    def test_get_mailman_client_hooks(self):

        @mailman.mailmanclient_request_hook
        def hook(params):
            return params

        # When not in debug mode, the hooks aren't added to Client.
        self.assertTrue(hook in mailman.get_request_hooks())
        self.assertEqual(len(mailman.get_request_hooks()), 1)
